import ldap
import ldap.sasl

import os
try:
    import kadm5
except:
    kadm5 = None
import mit_error
import heimdal_error

from DSGlobals import *
from ldap.filter import escape_filter_chars, filter_format
from DSNodeTypes import NodeFactory
import DSContainer
import socket

SECURITY_ANY = 0
SECURITY_NONE = 1
SECURITY_SSL = 2
SECURITY_TLS = 3


class dict_obj(object):
    """A wrapper to access a dictionary keys as object attributes"""
    def __init__(self, d):
        self.d = d
    def __getattr__(self, attr):
        return self.d[attr]


class LDAPConnWrapper(ldap.ldapobject.ReconnectLDAPObject):
    """A wrapper around the ldap object which catches SERVER_DOWN errors.

    Allows an exception handler to be specified so that exceptions thrown by
    the search, search_s etc. methods can be forwarded to the appropriate
    place.
    """
    def __init__(self, *args, **kwargs):
        self._handlers = {}
        self._exceptionMap = {'ldap.SERVER_DOWN': Disconnected}
        self._reverseMap = dict([(y,x) for x,y in self._exceptionMap.items()])
        ldap.ldapobject.ReconnectLDAPObject.__init__(self, *args, **kwargs)

    def RegisterHandler(self, event, callback):
        self._handlers[str(event)] = callback

    def _pass_exception_info(self, err):
        print "pass exception info: %s\n __class__:%s" % (err, err.__class__)
        if self._exceptionMap.has_key(str(err.__class__)):
            print "exception map has key"
            ex = self._exceptionMap[str(err.__class__)]
            print self._handlers
            if self._handlers.has_key(str(ex)):
                self._handlers[str(ex)](err)

    def search_s(self, *args, **kwargs):
        try:
            return ldap.ldapobject.ReconnectLDAPObject.search_s(self, *args, **kwargs)
        except Exception, err:
            self._pass_exception_info(err)
            raise err

    def search(self, *args, **kwargs):
        try:
            return ldap.ldapobject.ReconnectLDAPObject.search(self, *args, **kwargs)
        except Exception, err:
            self._pass_exception_info(err)
            raise err

    def add(self, *args, **kwargs):
        try:
            return ldap.ldapobject.ReconnectLDAPObject.add(self, *args, **kwargs)
        except Exception, err:
            self._pass_exception_info(err)
            raise err

    def result(self, *args, **kwargs):
        try:
            return ldap.ldapobject.ReconnectLDAPObject.result(self, *args, **kwargs)
        except Exception, err:
            self._pass_exception_info(err)
            raise err

ldap_init = LDAPConnWrapper
#if hasattr(ldap.ldapobject, 'ReconnectLDAPObject'):
#
#    ldap_init = ldap.ldapobject.ReconnectLDAPObject
#else:
#    ldap_init = ldap.initialize


class SearchContainer:
    """A virtual, DSContainer-like object which represents the result of a search instead of the contents of an ou"""
    def __init__(self, dsconn, name, query, description=None, implements=None):
        self.dsconn = dsconn
        self.name = name
        self.query = query
        self.description = description
        self.interface = implements

    def Name(self):
        return self.name
        
    def Path(self):
        return "name=%s,type=search" % (self.name)
    
    def getChildNodes(self):
        children = self.dsconn._search(self.query)
        if self.interface:
            children = [child for child in children \
                if child.Implements(self.interface)]
        return children
        

    def getChildContainers(self):
        return []

    def _raise_create_error(self):
        raise DSError, "Cannot create anything in a virtual folder"

    def CreateUser(self, *args, **kwargs):
        self._raise_create_error()

    def CreateGroup(self, *args, **kwargs):
        self._raise_create_error()

    def DeleteNode(self, node, *args, **kwargs):
        return node.parent.DeleteNode(node, *args, **kwargs)


class DSConnection:
    """Represents a connection to a directory server"""

    def TryAutoConnect(cls, passwordCallback):
        """Try to detect a DNS-specified ldap server and log in using kerberos, return either a new DSConnection object or None.
        
        Checks that:
            The user has kerberos tickets in the credential cache
            A connection to the kadmin server can be established
            The ldap server supports gssapi
            The ldap server accepts the user's credentials

        The kerberos libraries may need to prompt for a password, so a callback
        must be provided. The callback may indicate that the operation should
        be cancelled by returning None.
        """

        # We now only auto-connect if kerberos administration
        # will be taking place
        if (not kadm5):
            return None

        haveMIT = False
        haveHeimdal = False

        try:
            kadm5.init_libs(kadm5.KRB_FLAVOR_HEIMDAL)
        except OSError, err:
            print "info: unable to initialize Heimdal kerberos: %s" % (err)
        else:
            haveHeimdal = True

        try:
            kadm5.init_libs(kadm5.KRB_FLAVOR_MIT)
        except OSError, err:
            print "info: unable to initialize MIT kerberos: %s" % (err)
        else:
            haveMIT = True

        if not (haveMIT or haveHeimdal):
            print "No kerberos libs found"
            return None

        # This seems to work across kerberos flavors
        krb = kadm5.KRB5()
        try:
            principals = krb.klist()
        except Exception, err:
            print err
            return None

        if (not principals):
            return None

        # The first principal is the default.
        principal = principals[0]
        print "found principal: %s" % (principal)


        if (haveMIT and haveHeimdal):
            # We will try to detect which set of libs to use based on how
            # heimdal vs mit kerberos. This is pretty gross.
            
            kadm5.init_libs(kadm5.KRB_FLAVOR_HEIMDAL)
            k = kadm5.KADM5()
            k.connect(princ=principal)
                    
            try:
                k.GetPrincipal(principal)
            except heimdal_error.KADM5_AUTH_GET, err:
                print "info: kerberos account is unprivileged"
                return None
            except Exception, err:
                print err
                print "info: looks like an MIT kadmind..."
                haveHeimdal = False
            else:
                print "info: looks like a Heimdal kadmind..."

        kadmin = None
        
        if (haveHeimdal):
            kadm5.init_libs(kadm5.KRB_FLAVOR_HEIMDAL)
            kadmin = kadm5.KADM5()
            kadmin.connect(princ=principal)
            print "Connected to the heimdal kadmind"

        else:
            kadm5.init_libs(kadm5.KRB_FLAVOR_MIT)
            kadmin = kadm5.KADM5()

            useAdminAccount = True
            for passwordRetries in range(3):
                password = passwordCallback()
                if not password:
                    return None

                # We first try username/admin@REALM.
                # if no such principal exists, fall back to username@REALM
                try:
                    if (useAdminAccount):
                        try:
                            kadmin.connect(password=password)
                            break
                        except mit_error.KRB5KDC_ERR_C_PRINCIPAL_UNKNOWN, err:
                            print "info: no /admin account found"
                            useAdminAccount = False

                    if (not useAdminAccount):
                        kadmin.connect(
                                princ=principal,
                                password=password
                        )
                        break

                except mit_error.KADM5_BAD_PASSWORD:
                    continue
                except Exception, err:
                    print "kadm5 operations failed"
                    print err
                    return None
        
        #realm = cls._KerberosRealm()
        #detectedKrb5Hosts = cls._LocateDNS('_kerberos._udp')
        #detectedKadmHosts = cls._LocateDNS('_kerberos-adm._tcp')

        #if (not detectedKadmHosts):
        #    kadmHost, kadmPort = detectedKadmHosts[0]
        
        #hostname = socket.getfqdn()
        #if hostname
        
        # Locate LDAP Servers
        detectedHosts = cls._LocateDNS()

        
        if (detectedHosts):
            print "Got ldap server from DNS. Attempting kerberos login to",
            host, port = detectedHosts[0]
            print host, port
            try:
                conn, security = cls._do_connect(host, port)
            except DSError, err:
                print err
                return None
            try:
                gss = ldap.sasl.gssapi()
                conn.sasl_interactive_bind_s('', gss)
            except Exception, err:
                print err
                return None
            print "Connected with security type: %d, as: %s" % (security, conn.whoami_s())
            dsconn = cls(host, port, None, security)
            dsconn.conn = conn
            dsconn._gss_obj = gss
            #dsconn._realm = realm
            #dsconn._kadm_hosts = detectedKadmHosts
            #dsconn._krb5_hosts = detectedKrb5Hosts
            dsconn.kadmin = kadmin
            dsconn._discover_basedn()
            print dsconn.basedn
            return dsconn
        else:
            print "No LDAP servers auto-detected"

    TryAutoConnect = classmethod(TryAutoConnect) 

    def _do_connect(cls, host, port = None, security = SECURITY_ANY): 

        conn = None

        # Try each type of security, starting with most secure
        if (security == SECURITY_ANY):
            for sec in (SECURITY_TLS, SECURITY_SSL):
                try:
                    conn, security = cls._do_connect(host, port, sec)
                    break
                except Exception, err:
                    print "Info: Autodetecting; %s" % (err)
            if not (conn):
                # Neither of the secure methods succeeded
                # Connect normally so that descriptive errors
                # can be generated instead of a catch-all
                security = SECURITY_NONE
                conn, security = cls._do_connect(host, port, security)
        else:
            try:
                if (security == SECURITY_TLS):
                    if (port):
                        uri = "ldap://%s:%d" % (host, port)
                    else:  
                        uri = "ldap://%s" % (host)
                    conn = ldap_init(uri)
                    conn.start_tls_s()
                elif (security == SECURITY_SSL):
                    if (port):
                        uri = "ldaps://%s:%d" % (host, port)
                    else:
                        uri = "ldaps://%s" % (host)
                    conn = ldap_init(uri)
                    conn.search_s('', ldap.SCOPE_BASE)
                elif (security == SECURITY_NONE):
                    if (port):
                        uri = "ldap://%s:%d" % (host, port)
                    else:
                        uri = "ldap://%s" % (host)
                    conn = ldap_init(uri)
                    conn.search_s('', ldap.SCOPE_BASE)
                else:
                    raise DSError("Bad security type specified")

            except ldap.CONNECT_ERROR, err:
                raise ConnectError("Error in connection. Possible bad certificate chain.")
            except ldap.SERVER_DOWN, err:
                raise ConnectError("Error in connection. Can't contact LDAP server.")
            except Exception, err:
                raise DSError(err)

        return (conn, security)

    _do_connect = classmethod(_do_connect)

    def LocateHosts(cls):
        """Use DNS or DNS-SD to determine available ldap servers"""

        servers = []
        servers.extend(cls._LocateDNS())
        servers.extend(cls._LocateHowl())
        return servers
    LocateHosts = classmethod(LocateHosts)

    def _DNSQuery(cls, service='_ldap._tcp', qtype='srv'):
        try:
            import DNS
            
            # Try to get the domain name
            hostname = socket.gethostname()
            fqdn = socket.getfqdn()
            if not (fqdn.count('.')):
                return None
            domain = '.'.join(fqdn.split('.')[1:])

            DNS.DiscoverNameServers()
            req = DNS.Request(qtype='srv')
            res = req.req('%s.%s' % (service, domain))
            return list(res.answers)
        except:
            return []

    _DNSQuery = classmethod(_DNSQuery)
    
    def _LocateDNS(cls, service='_ldap._tcp'):
            
        reslist = cls._DNSQuery(service, 'srv')
        if not reslist:
            return []

        # Honor the priority given by DNS
        reslist.sort(lambda a, b: cmp(a['data'][0], b['data'][0]))

        # create a list of (host, port) tuples
        return [(r['data'][3], r['data'][2]) for r in reslist]

    _LocateDNS = classmethod(_LocateDNS)

    def _KerberosRealm(cls):
    
        reslist = cls._DNSQuery('_kerberos', 'txt')
        if not reslist:
            return None
        else:
            return reslist[0]['data'][0]
    _KerberosRealm = classmethod(_KerberosRealm)

    def _LocateHowl(cls):
        try:
            import howl

            D = howl.Discovery()
            res = D.Browse('_ldap._tcp', 500)
            # create a list of (host, port) tuples
            return [(r[3], r[4]) for r in res]
            
        except:
            return []
    _LocateHowl = classmethod(_LocateHowl)


    def __init__(self, host, port=None, basedn=None, security=SECURITY_ANY):
        """Create a DSConnection object.

        If port is not specified, the default (389) will be used. If basedn or
        security are not specified, they will be autodetected. Security is
        detected by trying first TLS, then SSL, then unencrypted connection
        types.
        """
        self.host = host
        self.username = None
        self.password = None
        self.port = port
        self.basedn = basedn
        self.security = security
        self.conn = None
        self.kadmin = None

    def connect(self, username, password):
        """Connect to the ldap server using simple username/password authentication.
        
        Either a dn or a username can be specified. If a username is specified,
        the directory is searched for a posixAccount matching 'uid=username'
        and its dn is used. A verbatim non-dn, non-posixAccount username may be
        used by putting it in quotes. This is useful esp. for active directory
        servers.
        """

        self.username = username
        self.password = password
        # initialize the ldap object using the specified security level
        self.conn, self.security = DSConnection._do_connect(self.host, self.port, self.security)

        # If no basedn was specified, attempt to determine the basedn
        if (not self.basedn):
            self._discover_basedn()

        # If a dn was given, use it. If only a username was specified,
        # find the associated dn
        if (self.username.count('=')):
            # the username is actually a dn
            self.binddn = self.username
        elif (self.username.startswith('"') and self.username.endswith('"')):
            # Allow users to enter e.g. "username@DOMAIN.COM" for non-DN
            # logins that should be used as-is
            self.binddn = self.username[1:-1]
        else:
            self._discover_binddn()

        try:
            self.conn.bind_s(self.binddn, self.password, ldap.AUTH_SIMPLE)
        except ldap.INVALID_CREDENTIALS, err:
            raise AuthError("Invalid username or password.")
        except ldap.INVALID_DN_SYNTAX, err:
            raise AuthError("Invalid username.")
        except Exception, err:
            raise DSError(err)


    def RegisterHandler(self, exc, func):
        return self.conn.RegisterHandler(exc, func)

    def _discover_basedn(self):
            resid = self.conn.search('', ldap.SCOPE_BASE, attrlist=['namingContexts'])
            try:
                rescode, results = self.conn.result(resid, LDAP_TIMEOUT)
            except ldap.TIMEOUT, err:
                raise Timeout("Timeout determining the rootdn")
        
            if (len(results) < 1):
                raise ConnectError("Couldn't determine the rootdn for the server")
            else:
                res = results[0]
                self.basedn = res[1]['namingContexts'][0]


    def _discover_binddn(self):
        # a dn was not provided, only a username
        # we'll do a search for the matching dn
        if (self.username == 'admin'):
            searchstring = '(cn=admin)'
            searchscope = ldap.SCOPE_ONELEVEL
        else:
            searchstring = filter_format('(uid=%s)', [self.username])
            searchscope = ldap.SCOPE_SUBTREE
            
        resid = self.conn.search(self.basedn, searchscope, searchstring, ['dn'])
        try:
            rescode, results = self.conn.result(resid, LDAP_TIMEOUT)
        except ldap.TIMEOUT, err:
            raise Timeout("Timeout determining the binddn")
    
        if (len(results) < 1):
            raise AuthError("Couldn't find an account for the provided username.")
        else:
            res = results[0]
            self.binddn = res[0]
       
    def getRootContainer(self):
        """Get a handle to the root DSContainer."""
        rootnodename = '.'.join(ldap.explode_dn(self.basedn, 1))
        bootstrapContainer = dict_obj({'conn' : self.conn,
                                    '_dsconn' : self })
        rootAttributes = (self.basedn, 
                          {'objectClass' : ['organizationalUnit'],
                           'ou' : [rootnodename]
                          }
                         )
        return NodeFactory(bootstrapContainer, rootAttributes)


    def _search(self, filter):

        resid = self.conn.search(self.basedn, ldap.SCOPE_SUBTREE, \
            filter, ('objectClass','displayName'))
        
        rescode, results = self.conn.result(resid)
        
        containers = {}
        nodes = []
        results.sort()
        dn_index = 0
        attrs_index = 1
        nodes = self._reslist_to_nodes(results)
        
        return nodes

    def GetObjects(self, paths, ignore_missing = False):
        """Return a list of nodes for the paths specified.

        If ignore_missing is specified, paths that point to non-existent
        objects will not cause the operation to fail, they will just be left
        out of the results.
        """
        ## Trying to create an easy api for retrieving a single
        ## Object by DN here. If only DN is given though, we'll
        ## Have to do a lookup just for that. So, that's not the
        ## "preferred" method to use. I wish you could do a search
        ## like (|(dn=...)(dn=...)) maybe you can?
        reslist = []
        for i in xrange(len(paths)):
            path = paths[i]
            # TODO: retrieval of objectClass goes here
            results = self.conn.search_s(path, ldap.SCOPE_BASE, \
                    attrlist=['objectClass', 'displayName'])
            # in a scope_base search, there can be only one result item
            # add that one item to the main reslist
            if (results):
                reslist.append(results[0])
            else:
                if not (ignore_missing):
                    raise DSError, "Couldn't retrieve the object with dn: %s" % (path)
        return self._reslist_to_nodes(reslist)

    def _reslist_to_nodes(self, reslist):
        "Return a list of nodes from a list of ldap result tuples"

        nodes = []
        containers = {}
        for res in reslist:
            path = res[0]
            parent_dn_parts = ldap.explode_dn(path)
            if (len(parent_dn_parts) < 2):
                raise DSError, "Invalid object path: %s" % (path)

            parent_dn = ','.join(parent_dn_parts[1:])
            parent_name = ldap.explode_dn(parent_dn, 1)[0]
            
            parent = containers.setdefault(parent_dn,
                NodeFactory(dict_obj({'conn':self.conn,'_dsconn':self}), 
                            (parent_dn, {'ou'          : [parent_name],
                                         'objectClass' : ['organizationalUnit']
                                        }
                            )
                           )
            )
            nodes.append(NodeFactory(parent, res))
        return nodes

    def GetObject(self, path):
        """Return the DSNode object for the given path."""
        ret = self.GetObjects([path])
        if (ret):
            return ret[0]
        else:
            return None
        
    def Groups(self):
        """Return a list of all the group nodes in the directory."""
        filter = "(objectClass=posixGroup)"
        return self._search(filter)

    def GroupsVFolder(self):
        """Return a virtual container representing all the group nodes in the directory."""
        filter = "(objectClass=posixGroup)"
        return SearchContainer(self, "All Groups", filter, implements="IDSGroup")

    def Users(self):
        """Return a list of all the user nodes in the directory."""
        filter = "(objectClass=posixAccount)"
        return self._search(filter)

    def UsersVFolder(self):
        """Return a virtual container representing all the user nodes in the directory."""
        filter = "(objectClass=posixAccount)"
        return SearchContainer(self, "All Users", filter, implements="IDSUser")

if __name__ == '__main__':
    from pprint import pprint
    print "Detecting LDAP servers..."
    hosts = DSConnection.LocateHosts()

    if (hosts):
        print "Detected hosts:"
        pprint (hosts)
        print "Connecting to host: %s..." % (hosts[0][0])
        conn = DSConnection(hosts[0][0], '', '', hosts[0][1])
        conn.connect()
        pprint (conn.getRootContainer())
